import json
import logging
from abc import ABC, abstractmethod
from json import JSONEncoder
from typing import List, Mapping, Set, Type

from dbcat.catalog import Catalog, CatColumn, CatSource, CatTable
from pglast import Node
from pglast.ast import RangeSubselect, RangeVar

from data_lineage import ColumnNotFound, SemanticError, TableNotFound
from data_lineage.parser.visitor import (
    ColumnRefVisitor,
    ExprVisitor,
    RangeSubselectVisitor,
    RangeVarVisitor,
)


class ColumnContext:
    def __init__(self, alias: str, columns: Set[CatColumn]):
        self._alias = alias.lower()
        self._columns = columns

    @property
    def alias(self):
        return self._alias

    @property
    def columns(self) -> Set[CatColumn]:
        return self._columns


class AliasContext:
    def __init__(self, catalog: Catalog, alias: str, tables: Set[CatTable]):
        self._catalog = catalog
        self._alias = alias.lower()
        self._tables = tables

    @property
    def alias(self):
        return self._alias

    @property
    def tables(self):
        return self._tables

    def get_columns(self, column_names: List[str] = None) -> List[ColumnContext]:
        columns: List[CatColumn] = []
        for table in self._tables:
            logging.debug("Searching in {}".format(table.fqdn))
            columns = columns + self._catalog.get_columns_for_table(table, column_names)

        return [
            ColumnContext(alias=column.name, columns={column}) for column in columns
        ]


class WithContext(AliasContext):
    def __init__(
        self,
        catalog: Catalog,
        alias: str,
        tables: Set[CatTable],
        columns: List[ColumnContext],
    ):
        super(WithContext, self).__init__(catalog, alias, tables)
        self._columns = columns

    def get_columns(self, column_names: List[str] = None) -> List[ColumnContext]:
        if column_names is not None:
            filtered = []
            for column in self._columns:
                logging.debug(
                    "Comparing with alias: {} - contains columns: {}".format(
                        column.alias,
                        json.dumps(list(column.columns), cls=CatColumnEncoder),
                    )
                )
                if column.alias in column_names:
                    filtered.append(column)

            return filtered
        else:
            return self._columns


class CatTableEncoder(JSONEncoder):
    def default(self, obj):
        if isinstance(obj, CatTable):
            return {
                "name": obj.name,
                "schema": obj.schema.name,
                "source": obj.schema.source.name,
            }

        # Let the base class default method raise the TypeError
        return json.JSONEncoder.default(self, obj)


class CatColumnEncoder(JSONEncoder):
    def default(self, obj):
        if isinstance(obj, CatColumn):
            return {
                "name": obj.name,
                "table": obj.table.name,
                "schema": obj.table.schema.name,
                "source": obj.table.schema.source.name,
            }

        # Let the base class default method raise the TypeError
        return json.JSONEncoder.default(self, obj)


class Binder(ABC):
    @property
    @abstractmethod
    def _visited_tables(self) -> List[Node]:
        pass

    @property
    @abstractmethod
    def _visited_columns(self) -> List[ExprVisitor]:
        pass

    @property
    def tables(self) -> Set[CatTable]:
        return self._tables

    @property
    def columns(self) -> List[ColumnContext]:
        return self._columns

    def __init__(
        self,
        catalog: Catalog,
        source: CatSource,
        alias_generator,
        expr_visitor_clazz: Type[ExprVisitor],
        alias_map: Mapping[str, AliasContext] = None,
    ):
        self._catalog = catalog
        self._source = source
        self._tables: Set[CatTable] = set()
        self._columns: List[ColumnContext] = []
        self._alias_map: Mapping[str, AliasContext] = alias_map or {}
        self._alias_generator = alias_generator
        self._expr_visitor_clazz = expr_visitor_clazz

    def bind(self):
        bound_tables = self._bind_tables()

        self._tables = set(bound_tables)
        self._columns = self._bind_columns()

    def _bind_tables(self):
        bound_tables = []
        for table in self._visited_tables:
            if isinstance(table, RangeVar):
                visitor = RangeVarVisitor()
                visitor(table)

                logging.debug("Searching for: {}".format(visitor.search_string))

                if not visitor.is_qualified and visitor.name in self._alias_map:
                    bound_tables = bound_tables + list(
                        self._alias_map[visitor.name].tables
                    )
                    logging.debug("Added tables for alias {}".format(visitor.name))
                else:
                    try:
                        candidate_table = self._catalog.search_table(
                            source_like=self._source.name, **visitor.search_string
                        )
                    except RuntimeError as err:
                        logging.debug(str(err))
                        raise TableNotFound(
                            '"{schema_like}"."{table_like}" is not found'.format(
                                **visitor.search_string
                            )
                        )
                    logging.debug("Bound source table: {}".format(candidate_table))

                    self._alias_map[visitor.alias] = AliasContext(
                        catalog=self._catalog,
                        alias=visitor.alias,
                        tables={candidate_table},
                    )
                    bound_tables.append(candidate_table)
            elif isinstance(table, RangeSubselect):
                visitor = RangeSubselectVisitor(self._expr_visitor_clazz)
                visitor(table)
                binder = SelectBinder(
                    self._catalog,
                    self._source,
                    visitor.sources,
                    visitor.columns,
                    self._alias_generator,
                    self._expr_visitor_clazz,
                )
                binder.bind()
                self._alias_map[visitor.alias] = WithContext(
                    catalog=self._catalog,
                    alias=visitor.alias,
                    tables=binder.tables,
                    columns=binder.columns,
                )
                bound_tables = bound_tables + list(binder.tables)
            else:
                raise SemanticError("Unknown parser state. Please contact Support")
        return bound_tables

    def _bind_columns(self) -> List[ColumnContext]:
        bound_cols: List[ColumnContext] = []
        for expr_visitor in self._visited_columns:
            target_cols: Set[ColumnContext] = set()
            is_a_star = False
            for column in expr_visitor.columns:
                column_ref_visitor = ColumnRefVisitor()
                column_ref_visitor(column)
                is_a_star = column_ref_visitor.is_a_star
                alias_list = list(self._alias_map.values())
                if column_ref_visitor.is_qualified:
                    if column_ref_visitor.table_name not in self._alias_map:
                        raise TableNotFound(
                            "{} not found for column ({}).".format(
                                column_ref_visitor.name[0], column_ref_visitor.name
                            )
                        )
                    assert column_ref_visitor.table_name is not None
                    alias_list = [self._alias_map[column_ref_visitor.table_name]]
                target_cols.update(
                    Binder._search_column_in_tables(column_ref_visitor, alias_list)
                )

            if is_a_star:
                for col in target_cols:
                    bound_cols.append(
                        ColumnContext(alias=col.alias, columns=col.columns)
                    )
            else:
                if expr_visitor.alias is not None:
                    alias = expr_visitor.alias
                elif len(target_cols) == 1:
                    alias = list(target_cols)[0].alias
                else:
                    alias = next(self._alias_generator)
                cols: Set[CatColumn] = set()
                for tgt in target_cols:
                    for c in tgt.columns:
                        cols.add(c)
                bound_cols.append(ColumnContext(alias=alias, columns=cols))

        if len(bound_cols) == 0:
            raise ColumnNotFound("No source columns found.")
        return bound_cols

    @staticmethod
    def _search_column_in_tables(
        column_ref_visitor, alias_list: List[AliasContext]
    ) -> List[ColumnContext]:
        found_cols: List[ColumnContext] = []
        if column_ref_visitor.is_a_star:
            for alias_context in alias_list:
                found_cols = alias_context.get_columns()
                logging.debug(
                    "Bound all source columns in {}".format(alias_context.tables)
                )
        else:
            candidate_columns: List[ColumnContext] = []
            global_table_list: List[CatTable] = []
            logging.debug("Searching for {}".format(column_ref_visitor.column_name))
            for alias_context in alias_list:
                logging.debug("Searching in {}".format(alias_context.alias))
                candidate_columns = candidate_columns + alias_context.get_columns(
                    [column_ref_visitor.column_name]
                )
                global_table_list = global_table_list + list(alias_context.tables)

            if len(candidate_columns) == 0:
                raise ColumnNotFound(
                    '"{}" not found in the following tables: {}'.format(
                        column_ref_visitor.column_name,
                        json.dumps(global_table_list, cls=CatTableEncoder),
                    )
                )
            elif len(candidate_columns) > 1:
                column_list = []
                for candidate in candidate_columns:
                    for col in candidate.columns:
                        column_list.append(col)
                raise ColumnNotFound(
                    "{} Ambiguous column name. Multiple matches found: {}".format(
                        column_ref_visitor.name,
                        json.dumps(column_list, cls=CatColumnEncoder),
                    )
                )
            logging.debug("Bound source column: {}".format(candidate_columns[0]))
            found_cols.append(candidate_columns[0])
        return found_cols


class SelectBinder(Binder):
    def __init__(
        self,
        catalog: Catalog,
        source: CatSource,
        tables: List[Node],
        columns: List[ExprVisitor],
        alias_generator,
        expr_visitor_clazz: Type[ExprVisitor],
        alias_map: Mapping[str, AliasContext] = None,
    ):
        super(SelectBinder, self).__init__(
            catalog, source, alias_generator, expr_visitor_clazz, alias_map
        )
        self._table_nodes: List[Node] = tables
        self._column_nodes: List[ExprVisitor] = columns

    @property
    def _visited_tables(self) -> List[Node]:
        return self._table_nodes

    @property
    def _visited_columns(self) -> List[ExprVisitor]:
        return self._column_nodes



import logging
from typing import List

from dbcat.catalog import Catalog
from dbcat.catalog.models import CatSource, JobExecution, JobExecutionStatus
from pglast import Node, parse_sql
from pglast.parser import ParseError

from data_lineage import SemanticError
from data_lineage.parser.binder import SelectBinder
from data_lineage.parser.dml_visitor import (
    CTASVisitor,
    DmlVisitor,
    SelectIntoVisitor,
    SelectSourceVisitor,
)
from data_lineage.parser.visitor import ExprVisitor, RedshiftExprVisitor


class Parsed:
    def __init__(self, name: str, query: str, node: Node):
        self._name = name
        self._node = node
        self._query = query

    @property
    def name(self):
        return self._name

    @property
    def node(self):
        return self._node

    @property
    def query(self):
        return self._query


def parse_queries(queries: List[str]) -> List[Parsed]:
    parsed: List[Parsed] = []

    for query in queries:
        try:
            parsed.append(parse(query))
        except ParseError as e:
            logging.warning("Syntax error while parsing {}.\n{}".format(query, e))

    return parsed


def analyze_dml_query(
    catalog: Catalog, parsed: Parsed, source: CatSource,
) -> DmlVisitor:
    chosen_visitor = visit_dml_query(parsed, source)
    chosen_visitor.bind(catalog=catalog, source=source)
    return chosen_visitor


def parse_dml_query(
    catalog: Catalog, parsed: Parsed, source: CatSource,
) -> SelectBinder:
    chosen_visitor = visit_dml_query(parsed, source)

    select_binder = SelectBinder(
        catalog=catalog,
        source=source,
        tables=chosen_visitor.select_tables,
        columns=chosen_visitor.select_columns,
        expr_visitor_clazz=chosen_visitor.expr_visitor_clazz,
        alias_generator=("_U{}".format(i) for i in range(0, 1000)),
    )
    select_binder.bind()
    return select_binder


def visit_dml_query(parsed: Parsed, source: CatSource,) -> DmlVisitor:

    expr_visitor_clazz = ExprVisitor
    if source.source_type == "redshift":
        expr_visitor_clazz = RedshiftExprVisitor

    select_source_visitor: DmlVisitor = SelectSourceVisitor(
        parsed.name, expr_visitor_clazz
    )
    select_into_visitor: DmlVisitor = SelectIntoVisitor(parsed.name, expr_visitor_clazz)
    ctas_visitor: DmlVisitor = CTASVisitor(parsed.name, expr_visitor_clazz)

    for v in [select_source_visitor, select_into_visitor, ctas_visitor]:
        v(parsed.node)
        if len(v.select_tables) > 0 and v.insert_table is not None:
            return v
    raise SemanticError("Query is not a DML Query")


def extract_lineage(
    catalog: Catalog,
    visited_query: DmlVisitor,
    source: CatSource,
    parsed: Parsed,
    start_time,
    end_time,
) -> JobExecution:
    job = catalog.add_job(
        name=parsed.name, source=source, context={"query": parsed.query}
    )
    job_execution = catalog.add_job_execution(
        job=job,
        started_at=start_time,
        ended_at=end_time,
        status=JobExecutionStatus.SUCCESS,
    )
    for source, target in zip(
        visited_query.source_columns, visited_query.target_columns
    ):
        for column in source.columns:
            edge = catalog.add_column_lineage(column, target, job_execution.id, {})
            logging.debug("Added {}".format(edge))

    return job_execution


def parse(sql: str, name: str = None) -> Parsed:
    if name is None:
        name = str(hash(sql))
    node = parse_sql(sql)

    return Parsed(name, sql, node)



from typing import Any, Dict, List, Optional, Tuple, Type

from pglast import Node
from pglast.visitors import Skip, Visitor


class ExprVisitor(Visitor):
    def __init__(self, alias: str = None):
        self._alias: Optional[str] = alias
        self._columns: List[Node] = []

    @property
    def alias(self) -> Optional[str]:
        return self._alias

    @property
    def columns(self) -> List[Node]:
        return self._columns

    def visit_FuncCall(self, ancestors, node):
        super().__call__(node.args)

    def visit_TypeCast(self, ancestors, node):
        super().__call__(node.arg)

    def visit_A_Expr(self, ancestors, node):
        super().__call__(node.lexpr)
        super().__call__(node.rexpr)

    def visit_ColumnRef(self, ancestors, node):
        self._columns.append(node)


class RedshiftExprVisitor(ExprVisitor):
    class FuncNameVisitor(Visitor):
        def __init__(self):
            self._name = None

        @property
        def name(self):
            return self._name

        def visit_String(self, ancestors, obj):
            self._name = obj.val

    def visit_FuncCall(self, ancestors, node):
        name_visitor = RedshiftExprVisitor.FuncNameVisitor()
        name_visitor(node.funcname)
        if name_visitor.name == "dateadd":
            super().__call__(node.args[2])
            return Skip


class TableVisitor(Visitor):
    def __init__(self, expr_visitor_clazz: Type[ExprVisitor]):
        self._sources: List[Node] = []
        self._columns: List[ExprVisitor] = []
        self._expr_visitor_clazz = expr_visitor_clazz
        self._with_aliases: Dict[str, Dict[str, Any]] = {}

    @property
    def sources(self) -> List[Node]:
        return self._sources

    @property
    def columns(self) -> List[ExprVisitor]:
        return self._columns

    @property
    def with_aliases(self) -> Dict[str, Dict[str, Any]]:
        return self._with_aliases

    def visit_ResTarget(self, ancestors, node):
        name = None
        if node.name is not None:
            name = node.name

        expr_visitor = self._expr_visitor_clazz(name)
        expr_visitor(node.val)
        self._columns.append(expr_visitor)
        return Skip

    def visit_RangeVar(self, ancestors, node):
        self._sources.append(node)
        return Skip

    def visit_RangeSubselect(self, ancestors, node):
        self._sources.append(node)
        return Skip

    def visit_CommonTableExpr(self, ancestors, node):
        with_alias = node.ctename
        table_visitor = TableVisitor(self._expr_visitor_clazz)
        table_visitor(node.ctequery)

        self._with_aliases[with_alias] = {
            "tables": table_visitor.sources,
            "columns": table_visitor.columns,
        }
        return Skip


class ColumnRefVisitor(Visitor):
    def __init__(self):
        self._name: List[str] = []
        self._is_a_star: bool = False

    @property
    def name(self) -> Tuple:
        return tuple(self._name)

    @property
    def is_a_star(self) -> bool:
        return self._is_a_star

    @property
    def is_qualified(self) -> bool:
        return len(self._name) == 2 or (len(self._name) == 1 and self._is_a_star)

    @property
    def column_name(self) -> Optional[str]:
        if len(self._name) == 2:
            return self._name[1]
        elif len(self._name) == 1:
            return self._name[0]
        return None

    @property
    def table_name(self) -> Optional[str]:
        if len(self._name) == 2 or (self._is_a_star and len(self._name) == 1):
            return self._name[0]

        return None

    def visit_String(self, ancestors, node):
        self._name.append(node.val)

    def visit_A_Star(self, ancestors, node):
        self._is_a_star = True


class RangeVarVisitor(Visitor):
    def __init__(self):
        self._schema_name = None
        self._name = None
        self._alias = None

    @property
    def alias(self) -> Optional[str]:
        if self._alias is not None:
            return self._alias
        elif self._schema_name is not None and self._name is not None:
            return "{}.{}".format(self._schema_name, self._name)
        elif self._name is not None:
            return self._name
        return None

    @property
    def fqdn(self):
        return self._schema_name, self._name

    @property
    def search_string(self):
        return {"schema_like": self._schema_name, "table_like": self._name}

    @property
    def is_qualified(self) -> bool:
        return self._schema_name is not None

    @property
    def schema_name(self) -> Optional[str]:
        return self._schema_name

    @property
    def name(self) -> str:
        return self._name

    def visit_Alias(self, ancestors, node):
        self._alias = node.aliasname.lower()

    def visit_RangeVar(self, ancestors, node):
        if node.schemaname:
            self._schema_name = node.schemaname.lower()
        self._name = node.relname.lower()


class RangeSubselectVisitor(Visitor):
    def __init__(self, expr_visitor_clazz: Type[ExprVisitor]):
        self._alias: Optional[str] = None
        self._table_visitor: TableVisitor = TableVisitor(expr_visitor_clazz)

    @property
    def alias(self) -> Optional[str]:
        if self._alias is not None:
            return self._alias
        return None

    @property
    def sources(self) -> List[Node]:
        return self._table_visitor.sources

    @property
    def columns(self) -> List[ExprVisitor]:
        return self._table_visitor.columns

    def visit_Alias(self, ancestors, node):
        self._alias = node.aliasname

    def visit_RangeSubselect(self, ancestors, node):
        super().__call__(node.alias)
        self._table_visitor(node.subquery)
        return Skip



import datetime
import logging
from typing import Any, Dict, List, Tuple

import flask_restless
import gunicorn.app.base
from dbcat import Catalog, PGCatalog, init_db
from dbcat.catalog import CatColumn
from dbcat.catalog.db import DbScanner
from dbcat.catalog.models import (
    CatSchema,
    CatSource,
    CatTable,
    ColumnLineage,
    DefaultSchema,
    Job,
    JobExecution,
    JobExecutionStatus,
)
from flask import Flask
from flask_restful import Api, Resource, reqparse
from pglast.parser import ParseError
from werkzeug.exceptions import NotFound, UnprocessableEntity

from data_lineage import ColumnNotFound, SemanticError, TableNotFound
from data_lineage.parser import (
    analyze_dml_query,
    extract_lineage,
    parse,
    parse_dml_query,
)


class TableNotFoundHTTP(NotFound):
    """Table not found in catalog"""

    code = 441


class ColumnNotFoundHTTP(NotFound):
    """Column not found in catalog"""

    code = 442


class ParseErrorHTTP(UnprocessableEntity):
    """Parser Error"""


class SemanticErrorHTTP(UnprocessableEntity):
    """Semantic Error"""

    code = 443


class Kedro(Resource):
    def __init__(self, catalog: Catalog):
        self._catalog = catalog
        self._parser = reqparse.RequestParser()
        self._parser.add_argument(
            "job_ids", action="append", help="List of job ids for a sub graph"
        )

    def get(self):
        nodes = []
        edges = []

        args = self._parser.parse_args()
        with self._catalog.managed_session:
            column_edges = self._catalog.get_column_lineages(args["job_ids"])
            for edge in column_edges:
                nodes.append(self._column_info(edge.source))
                nodes.append(self._column_info(edge.target))
                nodes.append(self._job_info(edge.job_execution.job))
                edges.append(
                    {
                        "source": "column:{}".format(edge.source_id),
                        "target": "task:{}".format(edge.job_execution.job_id),
                    }
                )
                edges.append(
                    {
                        "source": "task:{}".format(edge.job_execution.job_id),
                        "target": "column:{}".format(edge.target_id),
                    }
                )

        return {"nodes": nodes, "edges": edges}

    @staticmethod
    def _column_info(node: CatColumn):
        return {
            "id": "column:{}".format(node.id),
            "name": ".".join(node.fqdn),
            "type": "data",
        }

    @staticmethod
    def _job_info(node: Job):
        return {"id": "task:{}".format(node.id), "name": node.name, "type": "task"}


class Scanner(Resource):
    def __init__(self, catalog: Catalog):
        self._catalog = catalog
        self._parser = reqparse.RequestParser()
        self._parser.add_argument("id", required=True, help="ID of the resource")

    def post(self):
        args = self._parser.parse_args()
        logging.debug("Args for scanning: {}".format(args))
        with self._catalog.managed_session:
            source = self._catalog.get_source_by_id(int(args["id"]))
            DbScanner(self._catalog, source).scan()
            return "Scanned {}".format(source.fqdn), 200


class Parse(Resource):
    def __init__(self, catalog: Catalog):
        self._catalog = catalog
        self._parser = reqparse.RequestParser()
        self._parser.add_argument("query", required=True, help="Query to parse")
        self._parser.add_argument(
            "source_id", help="Source database of the query", required=True
        )

    def post(self):
        args = self._parser.parse_args()
        logging.debug("Parse query: {}".format(args["query"]))
        try:
            parsed = parse(args["query"], "parse_api")
        except ParseError as error:
            raise ParseErrorHTTP(description=str(error))

        try:
            with self._catalog.managed_session:
                source = self._catalog.get_source_by_id(args["source_id"])
                logging.debug("Parsing query for source {}".format(source))
                binder = parse_dml_query(
                    catalog=self._catalog, parsed=parsed, source=source
                )

                return (
                    {
                        "select_tables": [table.name for table in binder.tables],
                        "select_columns": [context.alias for context in binder.columns],
                    },
                    200,
                )
        except TableNotFound as table_error:
            raise TableNotFoundHTTP(description=str(table_error))
        except ColumnNotFound as column_error:
            raise ColumnNotFoundHTTP(description=str(column_error))
        except SemanticError as semantic_error:
            raise SemanticErrorHTTP(description=str(semantic_error))


class Analyze(Resource):
    def __init__(self, catalog: Catalog):
        self._catalog = catalog
        self._parser = reqparse.RequestParser()
        self._parser.add_argument("query", required=True, help="Query to parse")
        self._parser.add_argument("name", help="Name of the ETL job")
        self._parser.add_argument(
            "start_time", required=True, help="Start time of the task"
        )
        self._parser.add_argument(
            "end_time", required=True, help="End time of the task"
        )
        self._parser.add_argument(
            "source_id", help="Source database of the query", required=True
        )

    def post(self):
        args = self._parser.parse_args()
        logging.debug("Parse query: {}".format(args["query"]))
        try:
            parsed = parse(args["query"], args["name"])
        except ParseError as error:
            raise ParseErrorHTTP(description=str(error))

        try:
            with self._catalog.managed_session:
                source = self._catalog.get_source_by_id(args["source_id"])
                logging.debug("Parsing query for source {}".format(source))
                chosen_visitor = analyze_dml_query(self._catalog, parsed, source)
                job_execution = extract_lineage(
                    catalog=self._catalog,
                    visited_query=chosen_visitor,
                    source=source,
                    parsed=parsed,
                    start_time=datetime.datetime.fromisoformat(args["start_time"]),
                    end_time=datetime.datetime.fromisoformat(args["end_time"]),
                )

                return (
                    {
                        "data": {
                            "id": job_execution.id,
                            "type": "job_executions",
                            "attributes": {
                                "job_id": job_execution.job_id,
                                "started_at": job_execution.started_at.strftime(
                                    "%Y-%m-%d %H:%M:%S"
                                ),
                                "ended_at": job_execution.ended_at.strftime(
                                    "%Y-%m-%d %H:%M:%S"
                                ),
                                "status": job_execution.status.name,
                            },
                        }
                    },
                    200,
                )
        except TableNotFound as table_error:
            raise TableNotFoundHTTP(description=str(table_error))
        except ColumnNotFound as column_error:
            raise ColumnNotFoundHTTP(description=str(column_error))
        except SemanticError as semantic_error:
            raise SemanticErrorHTTP(description=str(semantic_error))


class Server(gunicorn.app.base.BaseApplication):
    def __init__(self, app, options=None):
        self.options = options or {}
        self.application = app
        super().__init__()

    def load_config(self):
        config = {
            key: value
            for key, value in self.options.items()
            if key in self.cfg.settings and value is not None
        }
        for key, value in config.items():
            self.cfg.set(key.lower(), value)

    def load(self):
        return self.application


def job_execution_serializer(instance: JobExecution, only: List[str]):
    return {
        "id": instance.id,
        "type": "job_executions",
        "attributes": {
            "job_id": instance.job_id,
            "started_at": instance.started_at.strftime("%Y-%m-%d %H:%M:%S"),
            "ended_at": instance.ended_at.strftime("%Y-%m-%d %H:%M:%S"),
            "status": instance.status.name,
        },
    }


def job_execution_deserializer(data: Dict["str", Any]):
    attributes = data["data"]["attributes"]
    logging.debug(attributes)
    job_execution = JobExecution()
    job_execution.job_id = int(attributes["job_id"])
    job_execution.started_at = datetime.datetime.strptime(
        attributes["started_at"], "%Y-%m-%d %H:%M:%S"
    )
    job_execution.ended_at = datetime.datetime.strptime(
        attributes["ended_at"], "%Y-%m-%d %H:%M:%S"
    )
    job_execution.status = (
        JobExecutionStatus.SUCCESS
        if attributes["status"] == "SUCCESS"
        else JobExecutionStatus.SUCCESS
    )

    logging.debug(job_execution)
    logging.debug(job_execution.status == JobExecutionStatus.SUCCESS)
    return job_execution


def create_server(
    catalog_options: Dict[str, str], options: Dict[str, str], is_production=True
) -> Tuple[Any, PGCatalog]:
    logging.debug(catalog_options)
    catalog = PGCatalog(
        **catalog_options,
        connect_args={"application_name": "data-lineage:flask-restless"},
        max_overflow=40,
        pool_size=20,
        pool_pre_ping=True
    )

    init_db(catalog)

    restful_catalog = PGCatalog(
        **catalog_options,
        connect_args={"application_name": "data-lineage:restful"},
        pool_pre_ping=True
    )

    app = Flask(__name__)

    # Create CRUD APIs
    methods = ["DELETE", "GET", "PATCH", "POST"]
    url_prefix = "/api/v1/catalog"
    api_manager = flask_restless.APIManager(app, catalog.get_scoped_session())
    api_manager.create_api(
        CatSource,
        methods=methods,
        url_prefix=url_prefix,
        additional_attributes=["fqdn"],
    )
    api_manager.create_api(
        CatSchema,
        methods=methods,
        url_prefix=url_prefix,
        additional_attributes=["fqdn"],
    )
    api_manager.create_api(
        CatTable,
        methods=methods,
        url_prefix=url_prefix,
        additional_attributes=["fqdn"],
    )
    api_manager.create_api(
        CatColumn,
        methods=methods,
        url_prefix=url_prefix,
        additional_attributes=["fqdn"],
    )
    api_manager.create_api(Job, methods=methods, url_prefix=url_prefix)
    api_manager.create_api(
        JobExecution,
        methods=methods,
        url_prefix=url_prefix,
        serializer=job_execution_serializer,
        deserializer=job_execution_deserializer,
    )
    api_manager.create_api(
        ColumnLineage,
        methods=methods,
        url_prefix=url_prefix,
        collection_name="column_lineage",
    )

    api_manager.create_api(
        DefaultSchema,
        methods=methods,
        url_prefix=url_prefix,
        collection_name="default_schema",
        primary_key="source_id",
    )

    restful_manager = Api(app)
    restful_manager.add_resource(
        Kedro, "/api/main", resource_class_kwargs={"catalog": restful_catalog}
    )
    restful_manager.add_resource(
        Scanner,
        "{}/scanner".format(url_prefix),
        resource_class_kwargs={"catalog": restful_catalog},
    )
    restful_manager.add_resource(
        Analyze, "/api/v1/analyze", resource_class_kwargs={"catalog": restful_catalog}
    )

    restful_manager.add_resource(
        Parse, "/api/v1/parse", resource_class_kwargs={"catalog": restful_catalog}
    )

    for rule in app.url_map.iter_rules():
        rule_methods = ",".join(rule.methods)
        logging.debug("{:50s} {:20s} {}".format(rule.endpoint, rule_methods, rule))

    if is_production:
        return Server(app=app, options=options), catalog
    else:
        return app, catalog



import logging
from typing import Dict, List

import networkx as nx


class LineageGraph:
    def __init__(
        self,
        nodes: List[Dict[str, str]],
        edges: List[Dict[str, str]],
        name: str = "Lineage",
    ):
        self.name = name
        self._graph = nx.DiGraph()
        for node in nodes:
            node_id = node["id"]
            node_attributes = {"name": node["name"], "type": node["type"]}
            logging.debug("Add Node: {}, {}".format(node_id, node_attributes))
            self._graph.add_node(node_id, **node_attributes)

        for edge in edges:
            logging.debug("Edge: <{}>, <{}>".format(edge["source"], edge["target"]))
            self._graph.add_edge(edge["source"], edge["target"])

    @property
    def graph(self):
        return self._graph

    @graph.setter
    def graph(self, new_graph):
        self._graph = new_graph



import logging

import click

from data_lineage import __version__
from data_lineage.server import create_server


@click.command()
@click.version_option(__version__)
@click.option(
    "-l", "--log-level", envvar="LOG_LEVEL", help="Logging Level", default="INFO"
)
@click.option(
    "--catalog-user", help="Database user name", envvar="CATALOG_USER", required=True
)
@click.option(
    "--catalog-password",
    help="Database Password",
    envvar="CATALOG_PASSWORD",
    required=True,
)
@click.option(
    "--catalog-host", help="Database Host", envvar="CATALOG_HOST", default="localhost"
)
@click.option(
    "--catalog-port", help="Database Password", envvar="CATALOG_PORT", default=5432
)
@click.option(
    "--catalog-db", help="Postgres Database", envvar="CATALOG_DB", default="tokern"
)
@click.option(
    "--server-address",
    help="The socket to bind to",
    envvar="SERVER_ADDRESS",
    default="127.0.0.1:4142",
)
@click.option(
    "--is-production/--not-production",
    help="Run server in development mode",
    default=True,
)
def main(
    log_level,
    catalog_user,
    catalog_password,
    catalog_host,
    catalog_port,
    catalog_db,
    server_address,
    is_production,
):
    logging.basicConfig(level=getattr(logging, log_level.upper()))
    catalog = {
        "user": catalog_user,
        "password": catalog_password,
        "host": catalog_host,
        "port": catalog_port,
        "database": catalog_db,
    }
    options = {"bind": server_address}
    app, catalog = create_server(catalog, options, is_production=is_production)
    if is_production:
        app.run()
    else:
        app.run(debug=True)


if __name__ == "__main__":
    main()



import json
import logging
from typing import Any, Dict, List, Optional, Set, Tuple, Type

from dbcat.catalog import Catalog, CatColumn, CatSource, CatTable
from pglast import Node
from pglast.ast import IntoClause
from pglast.visitors import Ancestor, Continue, Skip, Visitor

from data_lineage import ColumnNotFound, SemanticError, TableNotFound
from data_lineage.parser.binder import (
    CatTableEncoder,
    ColumnContext,
    SelectBinder,
    WithContext,
)
from data_lineage.parser.visitor import (
    ColumnRefVisitor,
    ExprVisitor,
    RangeVarVisitor,
    TableVisitor,
)


class DmlVisitor(Visitor):
    def __init__(self, name: str, expr_visitor_clazz: Type[ExprVisitor]):
        self._name = name
        self._insert_table: Optional[Node] = None
        self._insert_columns: List[str] = []
        self._target_table: Optional[CatTable] = None
        self._target_columns: List[CatColumn] = []
        self._source_tables: Set[CatTable] = set()
        self._source_columns: List[ColumnContext] = []
        self._select_tables: List[Node] = []
        self._select_columns: List[ExprVisitor] = []
        self._with_aliases: Dict[str, Dict[str, Any]] = {}
        self._alias_map: Dict[str, WithContext] = {}
        self._column_alias_generator = ("_U{}".format(i) for i in range(0, 1000))
        self.expr_visitor_clazz = expr_visitor_clazz

    @property
    def name(self) -> str:
        return self._name

    @property
    def insert_table(self) -> Optional[Node]:
        return self._insert_table

    @property
    def target_table(self) -> CatTable:
        return self._target_table

    @property
    def target_columns(self) -> List[CatColumn]:
        return self._target_columns

    @property
    def source_tables(self) -> Set[CatTable]:
        return self._source_tables

    @property
    def source_columns(self) -> List[ColumnContext]:
        return self._source_columns

    @property
    def select_tables(self) -> List[Node]:
        return self._select_tables

    @property
    def select_columns(self) -> List[ExprVisitor]:
        return self._select_columns

    def visit_RangeVar(self, ancestors, node):
        self._insert_table = node
        return Skip

    def visit_ResTarget(self, ancestors, node):
        self._insert_columns.append(node.name)
        return Skip

    def visit_CommonTableExpr(self, ancestors, node):
        with_alias = node.ctename
        table_visitor = TableVisitor(self.expr_visitor_clazz)
        table_visitor(node.ctequery)

        self._with_aliases[with_alias] = {
            "tables": table_visitor.sources,
            "columns": table_visitor.columns,
        }
        return Skip

    def visit_CreateTableAsStmt(self, ancestors, node):
        """
            Do not process CTAS statement by default.
        :param ancestors:
        :type ancestors:
        :param node:
        :type node:
        :return:
        :rtype:
        """
        return Skip

    def bind(self, catalog: Catalog, source: CatSource):
        self._bind_target(catalog, source)

        self._bind_with(catalog, source)
        binder = SelectBinder(
            catalog,
            source,
            self._select_tables,
            self._select_columns,
            self._column_alias_generator,
            self.expr_visitor_clazz,
            self._alias_map,
        )
        binder.bind()

        if len(binder.tables) == 0:
            raise SemanticError("No source tables found")

        if len(binder.columns) == 0:
            raise SemanticError("No source columns found")

        if self.target_table is None:
            raise SemanticError("No target table found")

        if len(self.target_columns) == 0:
            raise SemanticError(
                "No target columns found in {}".format(
                    json.dumps(self.target_table, cls=CatTableEncoder)
                )
            )

        if len(self.target_columns) != len(binder.columns):
            raise SemanticError(
                "No. of target columns({}) does not match no. of source columns({})".format(
                    len(self.target_columns), len(binder.columns)
                )
            )

        self._source_tables = binder.tables
        self._source_columns = binder.columns

    def _bind_target(self, catalog: Catalog, source: CatSource):
        target_table_visitor = RangeVarVisitor()
        target_table_visitor(self._insert_table)
        logging.debug("Searching for: {}".format(target_table_visitor.search_string))
        try:
            self._target_table = catalog.search_table(
                source_like=source.name, **target_table_visitor.search_string
            )
        except RuntimeError as error:
            logging.debug(str(error))
            raise TableNotFound(
                '"{schema_like}"."{table_like}" is not found'.format(
                    **target_table_visitor.search_string
                )
            )
        logging.debug("Bound target table: {}".format(self._target_table))
        if len(self._insert_columns) == 0:
            self._target_columns = catalog.get_columns_for_table(self._target_table)
            logging.debug("Bound all columns in {}".format(self._target_table))
        else:
            bound_cols = catalog.get_columns_for_table(
                self._target_table, column_names=self._insert_columns
            )
            # Handle error case
            if len(bound_cols) != len(self._insert_columns):
                for column in self._insert_columns:
                    found = False
                    for bound in bound_cols:
                        if column == bound.name:
                            found = True
                            break

                    if not found:
                        raise ColumnNotFound(
                            '"{}" not found in the following tables: {}'.format(
                                column,
                                json.dumps([self._target_table], cls=CatTableEncoder),
                            )
                        )

            self._target_columns = bound_cols
            logging.debug("Bound {} target columns".format(len(bound_cols)))

    def _bind_with(self, catalog: Catalog, source: CatSource):
        if self._with_aliases:
            # Bind all the WITH expressions
            for key in self._with_aliases.keys():
                binder = SelectBinder(
                    catalog,
                    source,
                    self._with_aliases[key]["tables"],
                    self._with_aliases[key]["columns"],
                    self._column_alias_generator,
                    self.expr_visitor_clazz,
                )
                binder.bind()
                self._alias_map[key] = WithContext(
                    catalog=catalog,
                    alias=key,
                    tables=binder.tables,
                    columns=binder.columns,
                )

    def resolve(
        self,
    ) -> Tuple[
        Tuple[Optional[str], str],
        List[Tuple[Optional[str], str]],
        List[Tuple[Optional[str], str]],
    ]:
        target_table_visitor = RangeVarVisitor()
        target_table_visitor(self._insert_table)

        bound_tables = []
        for table in self._select_tables:
            visitor = RangeVarVisitor()
            visitor(table)
            bound_tables.append(visitor.fqdn)

        bound_cols = []
        for expr_visitor in self._select_columns:
            for column in expr_visitor.columns:
                column_ref_visitor = ColumnRefVisitor()
                column_ref_visitor(column)
                bound_cols.append(column_ref_visitor.name[0])

        return target_table_visitor.fqdn, bound_tables, bound_cols


class SelectSourceVisitor(DmlVisitor):
    def __init__(self, name: str, expr_visitor_clazz: Type[ExprVisitor] = ExprVisitor):
        super(SelectSourceVisitor, self).__init__(name, expr_visitor_clazz)

    def visit_SelectStmt(self, ancestors, node):
        table_visitor = TableVisitor(self.expr_visitor_clazz)
        table_visitor(node)
        self._select_tables = table_visitor.sources
        self._select_columns = table_visitor.columns
        for key in table_visitor.with_aliases.keys():
            self._with_aliases[key] = table_visitor.with_aliases[key]

        return Skip


class SelectIntoVisitor(DmlVisitor):
    def __init__(self, name: str, expr_visitor_clazz: Type[ExprVisitor] = ExprVisitor):
        super(SelectIntoVisitor, self).__init__(name, expr_visitor_clazz)

    def visit_SelectStmt(self, ancestors, node):
        super().__call__(node.intoClause)
        table_visitor = TableVisitor(self.expr_visitor_clazz)
        table_visitor(node.targetList)
        table_visitor(node.fromClause)
        self._select_tables = table_visitor.sources
        self._select_columns = table_visitor.columns
        for key in table_visitor.with_aliases.keys():
            self._with_aliases[key] = table_visitor.with_aliases[key]

        return Skip


class CTASVisitor(SelectSourceVisitor):
    def __init__(self, name: str, expr_visitor_clazz: Type[ExprVisitor] = ExprVisitor):
        super(CTASVisitor, self).__init__(name, expr_visitor_clazz)

    def visit_CreateTableAsStmt(self, ancestors, node):
        return Continue

    def visit_String(self, ancestors: Ancestor, node):
        # Check if parent is IntoClause
        parent = ancestors
        in_into_clause = False
        while parent is not None and not in_into_clause:
            in_into_clause = isinstance(parent.node, IntoClause)
            parent = parent.parent

        if in_into_clause:
            self._insert_columns.append(node.val)

    def _bind_target(self, catalog: Catalog, source: CatSource):
        target_table_visitor = RangeVarVisitor()
        target_table_visitor(self._insert_table)

        if target_table_visitor.is_qualified:
            schema = catalog.get_schema(
                source_name=source.name, schema_name=target_table_visitor.schema_name
            )
        elif source.default_schema is not None:
            schema = source.default_schema.schema
        else:
            raise SemanticError(
                "No default schema set for source {}".format(source.fqdn)
            )

        self._target_table = catalog.add_table(
            table_name=target_table_visitor.name, schema=schema
        )

        sort_order = 1
        for col in self._insert_columns:
            self._target_columns.append(
                catalog.add_column(
                    column_name=col,
                    data_type="varchar",
                    sort_order=sort_order,
                    table=self._target_table,
                )
            )



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['data_lineage', 'data_lineage.parser']

package_data = \
{'': ['*'], 'data_lineage': ['assets/*']}

install_requires = \
['PyYAML',
 'click',
 'dbcat>=0.5.2',
 'flask-restful',
 'flask-restless-ng',
 'flask>=1.1,<2.0',
 'furl',
 'gunicorn',
 'inflection',
 'networkx',
 'pglast',
 'psycopg2>=2.9.1,<3.0.0',
 'requests']

entry_points = \
{'console_scripts': ['data_lineage = data_lineage.__main__:main']}

setup_kwargs = {
    'name': 'data-lineage',
    'version': '0.8.1',
    'description': 'Open Source Data Lineage Tool for Redshift. Snowflake and many other databases',
    'long_description': '# Tokern Lineage Engine\n\n[![CircleCI](https://circleci.com/gh/tokern/data-lineage.svg?style=svg)](https://circleci.com/gh/tokern/data-lineage)\n[![codecov](https://codecov.io/gh/tokern/data-lineage/branch/master/graph/badge.svg)](https://codecov.io/gh/tokern/data-lineage)\n[![PyPI](https://img.shields.io/pypi/v/data-lineage.svg)](https://pypi.python.org/pypi/data-lineage)\n[![image](https://img.shields.io/pypi/l/data-lineage.svg)](https://pypi.org/project/data-lineage/)\n[![image](https://img.shields.io/pypi/pyversions/data-lineage.svg)](https://pypi.org/project/data-lineage/)\n\n\nTokern Lineage Engine is _fast_ and _easy to use_ application to collect, visualize and analyze \ncolumn-level data lineage in databases, data warehouses and data lakes in AWS and GCP.\n\nTokern Lineage helps you browse column-level data lineage \n* visually using [kedro-viz](https://github.com/quantumblacklabs/kedro-viz)\n* analyze lineage graphs programmatically using the powerful [networkx graph library](https://networkx.org/)\n\n## Resources\n\n* Demo of Tokern Lineage App\n\n![data-lineage](https://user-images.githubusercontent.com/1638298/118261607-688a7100-b4d1-11eb-923a-5d2407d6bd8d.gif)\n\n* Checkout an [example data lineage notebook](http://tokern.io/docs/data-lineage/example/).\n\n* Check out [the post on using data lineage for cost control](https://tokern.io/blog/data-lineage-on-redshift/) for an \nexample of how data lineage can be used in production.\n\n## Quick Start\n\n### Install a demo of using Docker and Docker Compose\n\nDownload the docker-compose file from Github repository.\n\n\n    # in a new directory run\n    wget https://raw.githubusercontent.com/tokern/data-lineage/master/install-manifests/docker-compose/catalog-demo.yml\n    # or run\n    curl https://raw.githubusercontent.com/tokern/data-lineage/master/install-manifests/docker-compose/tokern-lineage-engine.yml -o docker-compose.yml\n\n\nRun docker-compose\n   \n\n    docker-compose up -d\n\n\nCheck that the containers are running.\n\n\n    docker ps\n    CONTAINER ID   IMAGE                                    CREATED        STATUS       PORTS                    NAMES\n    3f4e77845b81   tokern/data-lineage-viz:latest   ...   4 hours ago    Up 4 hours   0.0.0.0:8000->80/tcp     tokern-data-lineage-visualizer\n    1e1ce4efd792   tokern/data-lineage:latest       ...   5 days ago     Up 5 days                             tokern-data-lineage\n    38be15bedd39   tokern/demodb:latest             ...   2 weeks ago    Up 2 weeks                            tokern-demodb\n\nTry out Tokern Lineage App\n\nHead to `http://localhost:8000/` to open the Tokern Lineage app\n\n### Install Tokern Lineage Engine\n\n    # in a new directory run\n    wget https://raw.githubusercontent.com/tokern/data-lineage/master/install-manifests/docker-compose/tokern-lineage-engine.yml\n    # or run\n    curl https://raw.githubusercontent.com/tokern/data-lineage/master/install-manifests/docker-compose/catalog-demo.yml -o tokern-lineage-engine.yml\n\nRun docker-compose\n   \n\n    docker-compose up -d\n\n\nIf you want to use an external Postgres database, change the following parameters in `tokern-lineage-engine.yml`:\n\n* CATALOG_HOST\n* CATALOG_USER\n* CATALOG_PASSWORD\n* CATALOG_DB\n\nYou can also override default values using environement variables. \n\n    CATALOG_HOST=... CATALOG_USER=... CATALOG_PASSWORD=... CATALOG_DB=... docker-compose -f ... up -d\n\nFor more advanced usage of environment variables with docker-compose, [refer to docker-compose docs](https://docs.docker.com/compose/environment-variables/)\n\n**Pro-tip**\n\nIf you want to connect to a database in the host machine, set \n\n    CATALOG_HOST: host.docker.internal # For mac or windows\n    #OR\n    CATALOG_HOST: 172.17.0.1 # Linux\n\n## Supported Technologies\n\n* Postgres\n* AWS Redshift\n* Snowflake\n\n### Coming Soon\n\n* SparkSQL\n* Presto\n\n## Documentation\n\nFor advanced usage, please refer to [data-lineage documentation](https://tokern.io/docs/data-lineage/index.html)\n## Survey\n\nPlease take this [survey](https://forms.gle/p2oEQBJnpEguhrp3A) if you are a user or considering using data-lineage. Responses will help us prioritize features better. \n',
    'author': 'Tokern',
    'author_email': 'info@tokern.io',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://tokern.io/',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)



# flake8: noqa
__version__ = "0.8.1"

import datetime
import json
import logging
from typing import Any, Dict, Generator, Generic, List, Optional, Type, TypeVar

import requests
from dbcat.catalog.models import JobExecutionStatus
from furl import furl
from requests import HTTPError

from data_lineage.graph import LineageGraph


class SourceNotFound(Exception):
    """Source not found in catalog"""


class SchemaNotFound(Exception):
    """Schema not found in catalog"""


class TableNotFound(Exception):
    """Table not found in catalog"""


class ColumnNotFound(Exception):
    """Column not found in catalog"""


class ParseError(Exception):
    """Parser Error"""


class SemanticError(Exception):
    """Error due to mismatch in catalog data"""


class NoResultFound(Exception):
    """Raised when function returns no results"""


class MultipleResultsFound(Exception):
    """Raised when multiple results are found but expected only one or zero results"""


class Graph:
    def __init__(self, url: str):
        self._base_url = furl(url) / "api/main"
        self._session = requests.Session()

    def get(self, job_ids: set = None) -> Dict[str, List[Dict[str, str]]]:
        if job_ids is not None:
            response = self._session.get(
                self._base_url, params={"job_ids": list(job_ids)}
            )
        else:
            response = self._session.get(self._base_url)
        return response.json()


def load_graph(graphSDK: Graph, job_ids: set = None) -> LineageGraph:
    data = graphSDK.get(job_ids)
    return LineageGraph(nodes=data["nodes"], edges=data["edges"])


class BaseModel:
    def __init__(self, session, attributes, obj_id, relationships):
        self._session = session
        self._attributes = attributes
        self._obj_id = obj_id
        self._relationships = relationships

    def __getattr__(self, item):
        logging.debug("Attributes: {}".format(self._attributes))
        if item == "id":
            return self._obj_id
        elif self._attributes and item in self._attributes.keys():
            return self._attributes[item]
        elif self._relationships and item in self._relationships.keys():
            return self._relationships[item]
        raise AttributeError


class Source(BaseModel):
    def __init__(self, session, attributes, obj_id, relationships):
        super().__init__(session, attributes, obj_id, relationships)


class Schema(BaseModel):
    def __init__(self, session, attributes, obj_id, relationships):
        super().__init__(session, attributes, obj_id, relationships)


class Table(BaseModel):
    def __init__(self, session, attributes, obj_id, relationships):
        super().__init__(session, attributes, obj_id, relationships)


class Column(BaseModel):
    def __init__(self, session, attributes, obj_id, relationships):
        super().__init__(session, attributes, obj_id, relationships)


class Job(BaseModel):
    def __init__(self, session, attributes, obj_id, relationships):
        super().__init__(session, attributes, obj_id, relationships)


class JobExecution(BaseModel):
    def __init__(self, session, attributes, obj_id, relationships):
        super().__init__(session, attributes, obj_id, relationships)


class ColumnLineage(BaseModel):
    def __init__(self, session, attributes, obj_id, relationships):
        super().__init__(session, attributes, obj_id, relationships)


class DefaultSchema(BaseModel):
    def __init__(self, session, attributes, obj_id, relationships):
        super().__init__(session, attributes, obj_id, relationships)


ModelType = TypeVar("ModelType", bound=BaseModel)


class Catalog:
    def __init__(self, url: str):
        self._base_url = furl(url) / "api/v1/catalog"
        self._session = requests.Session()
        self._session.headers.update({"Accept": "application/vnd.api+json"})
        self._session.headers.update({"Content-Type": "application/vnd.api+json"})

    def _build_url(self, *urls) -> str:
        built_url = self._base_url
        for url in urls:
            built_url = furl(built_url) / url
        logging.debug(built_url)
        return built_url

    str_to_type = {
        "sources": Source,
        "schemata": Schema,
    }

    def _resolve_relationships(self, relationships) -> Dict[str, BaseModel]:
        resolved: Dict[str, BaseModel] = {}
        for key, value in relationships.items():
            logging.debug("Resolving {}:{}".format(key, value))
            if value["data"]:
                resolved[key] = self._obj_factory(
                    value["data"],
                    Catalog.str_to_type[value["data"]["type"]],
                    resolve_relationships=False,
                )

        return resolved

    def _obj_factory(
        self,
        payload: Dict[str, Any],
        clazz: Type[ModelType],
        resolve_relationships=False,
    ) -> ModelType:
        resolved = None
        if resolve_relationships and payload.get("relationships"):
            resolved = self._resolve_relationships(payload.get("relationships"))

        return clazz(
            session=self._session,
            attributes=payload.get("attributes"),
            obj_id=payload.get("id"),
            relationships=resolved,
        )

    def _iterate(self, payload: Dict[str, Any], clazz: Type[BaseModel]):
        res: Optional[Dict[str, Any]] = payload
        while res is not None:
            for item in res["data"]:
                yield self._obj_factory(payload=item, clazz=clazz)

            if res["links"]["next"] is not None:
                response = self._session.get(res["links"]["next"])
                res = response.json()
            else:
                res = None

    def _index(self, path: str, clazz: Type[BaseModel]):
        response = self._session.get(self._build_url(path))
        logging.debug(response.json())
        return self._iterate(response.json(), clazz)

    def _get(
        self,
        path: str,
        obj_id: int,
        clazz: Type[ModelType],
        resolve_relationships=False,
    ) -> ModelType:
        response = self._session.get(self._build_url(path, str(obj_id)))
        json_response = response.json()
        logging.debug(json_response)
        response.raise_for_status()
        return self._obj_factory(
            json_response["data"], clazz, resolve_relationships=resolve_relationships
        )

    @staticmethod
    def _one(response):
        json_response = response.json()
        logging.debug(json_response)
        num_results = json_response["meta"]["total"]
        if num_results == 0:
            raise NoResultFound
        elif num_results > 1:
            raise MultipleResultsFound

        return json_response["data"][0]

    def _search_one(self, path: str, filters):
        params = {"filter[objects]": json.dumps(filters)}
        response = self._session.get(self._build_url(path), params=params)
        response.raise_for_status()
        return Catalog._one(response)

    def _search(self, path: str, search_string: str, clazz: Type[BaseModel]):
        filters = [dict(name="name", op="like", val="%{}%".format(search_string))]
        params = {"filter[objects]": json.dumps(filters)}
        response = self._session.get(self._build_url(path), params=params)
        return self._iterate(response.json(), clazz)

    def _post(self, path: str, data: Dict[str, Any], type: str) -> Dict[Any, Any]:
        payload = {"data": {"type": type, "attributes": data}}
        response = self._session.post(
            url=self._build_url(path), data=json.dumps(payload, default=str)
        )
        response.raise_for_status()
        logging.debug(response.text)
        json_response = response.json()
        return json_response["data"]

    def _patch(self, path: str, obj_id: int, data: Dict[str, Any], type: str):
        payload = {"data": {"type": type, "attributes": data, "id": obj_id}}
        response = self._session.patch(
            url=self._build_url(path, str(obj_id)),
            data=json.dumps(payload, default=str),
        )
        response.raise_for_status()
        return

    def get_sources(self) -> Generator[Any, Any, None]:
        return self._index("sources", Source)

    def get_schemata(self):
        return self._index("schemata", Schema)

    def get_tables(self):
        return self._index("tables", Table)

    def get_columns(self):
        return self._index("columns", Column)

    def get_jobs(self):
        return self._index("jobs", Job)

    def get_job_executions(self):
        return self._index("job_executions", JobExecution)

    def get_column_lineages(self):
        return self._index("column_lineages", ColumnLineage)

    def get_source_by_id(self, obj_id) -> Source:
        return self._get("sources", obj_id, Source)

    def get_schema_by_id(self, obj_id) -> Schema:
        return self._get("schemata", obj_id, Schema)

    def get_table_by_id(self, obj_id) -> Table:
        return self._get("tables", obj_id, Table)

    def get_column_by_id(self, obj_id) -> Column:
        return self._get("columns", obj_id, Column)

    def get_job_by_id(self, obj_id) -> Job:
        return self._get("jobs", obj_id, Job)

    def get_job_execution_by_id(self, obj_id) -> JobExecution:
        return self._get("job_executions", obj_id, JobExecution)

    def get_column_lineage(self, job_ids: List[int]) -> List[ColumnLineage]:
        params = {"job_ids": job_ids}
        response = self._session.get(self._build_url("column_lineage"), params=params)
        logging.debug(response.json())
        response.raise_for_status()
        return [
            ColumnLineage(
                session=self._session,
                attributes=item["attributes"],
                obj_id=item["id"],
                relationships=item["relationships"],
            )
            for item in response.json()["data"]
        ]

    def get_source(self, name) -> Source:
        filters = [dict(name="name", op="eq", val="{}".format(name))]
        try:
            payload = self._search_one("sources", filters)
        except NoResultFound:
            raise SourceNotFound("Source not found: source_name={}".format(name))

        return self._obj_factory(payload, Source)

    def get_schema(self, source_name: str, schema_name: str) -> Schema:
        name_filter = dict(name="name", op="eq", val=schema_name)
        source_filter = dict(
            name="source", op="has", val=dict(name="name", op="eq", val=source_name)
        )
        filters = {"and": [name_filter, source_filter]}
        logging.debug(filters)
        try:
            payload = self._search_one("schemata", [filters])
        except NoResultFound:
            raise SchemaNotFound(
                "Schema not found, (source_name={}, schema_name={})".format(
                    source_name, schema_name
                )
            )
        return self._obj_factory(payload, Schema)

    def get_table(self, source_name: str, schema_name: str, table_name: str) -> Table:
        schema = self.get_schema(source_name, schema_name)

        name_filter = dict(name="name", op="eq", val=table_name)
        schema_id_filter = dict(name="schema_id", op="eq", val=str(schema.id))
        filters = {"and": [name_filter, schema_id_filter]}
        logging.debug(filters)
        try:
            payload = self._search_one("tables", [filters])
        except NoResultFound:
            raise TableNotFound(
                "Table not found, (source_name={}, schema_name={}, table_name={})".format(
                    source_name, schema_name, table_name
                )
            )
        return self._obj_factory(payload, Table)

    def get_columns_for_table(self, table: Table):
        return self._index("tables/{}/columns".format(table.id), Column)

    def get_column(self, source_name, schema_name, table_name, column_name) -> Column:
        table = self.get_table(source_name, schema_name, table_name)
        name_filter = dict(name="name", op="eq", val=column_name)
        table_filter = dict(name="table_id", op="eq", val=str(table.id))
        filters = {"and": [name_filter, table_filter]}
        logging.debug(filters)
        try:
            payload = self._search_one("columns", [filters])
        except NoResultFound:
            raise ColumnNotFound(
                "Column not found, (source_name={}, schema_name={}, table_name={}, column_name={})".format(
                    source_name, schema_name, table_name, column_name
                )
            )
        return self._obj_factory(payload, Column)

    def add_source(self, name: str, source_type: str, **kwargs) -> Source:
        data = {"name": name, "source_type": source_type, **kwargs}
        payload = self._post(path="sources", data=data, type="sources")
        return self._obj_factory(payload, Source)

    def scan_source(self, source: Source) -> bool:
        payload = {"id": source.id}
        response = self._session.post(
            url=self._build_url("scanner"), data=json.dumps(payload)
        )
        response.raise_for_status()
        return response.status_code == 200

    def add_schema(self, name: str, source: Source) -> Schema:
        data = {"name": name, "source_id": source.id}
        payload = self._post(path="schemata", data=data, type="schemata")
        return self._obj_factory(payload, Schema)

    def add_table(self, name: str, schema: Schema) -> Table:
        data = {"name": name, "schema_id": schema.id}
        payload = self._post(path="tables", data=data, type="tables")
        return self._obj_factory(payload, Table)

    def add_column(
        self, name: str, data_type: str, sort_order: int, table: Table
    ) -> Column:
        data = {
            "name": name,
            "table_id": table.id,
            "data_type": data_type,
            "sort_order": sort_order,
        }
        payload = self._post(path="columns", data=data, type="columns")
        return self._obj_factory(payload, Column)

    def add_job(self, name: str, context: Dict[Any, Any]) -> Job:
        data = {"name": name, "context": context}
        payload = self._post(path="jobs", data=data, type="jobs")
        return self._obj_factory(payload, Job)

    def add_job_execution(
        self,
        job: Job,
        started_at: datetime.datetime,
        ended_at: datetime.datetime,
        status: JobExecutionStatus,
    ) -> JobExecution:
        data = {
            "job_id": job.id,
            "started_at": started_at,
            "ended_at": ended_at,
            "status": status.name,
        }
        payload = self._post(path="job_executions", data=data, type="job_executions")
        return self._obj_factory(payload, JobExecution)

    def add_column_lineage(
        self,
        source: Column,
        target: Column,
        job_execution_id: int,
        context: Dict[Any, Any],
    ) -> ColumnLineage:
        data = {
            "source_id": source.id,
            "target_id": target.id,
            "job_execution_id": job_execution_id,
            "context": context,
        }
        payload = self._post(path="column_lineage", data=data, type="column_lineage")
        return self._obj_factory(payload, ColumnLineage)

    def update_source(self, source: Source, schema: Schema) -> DefaultSchema:
        try:
            current_obj = self._get(
                path="default_schema",
                obj_id=source.id,
                clazz=DefaultSchema,
                resolve_relationships=True,
            )
            if current_obj.schema.id == schema.id:
                return current_obj
        except HTTPError as error:
            if error.response.status_code == 404:
                data = {"source_id": source.id, "schema_id": schema.id}
                payload = self._post(
                    path="default_schema", data=data, type="default_schema"
                )
                return self._obj_factory(
                    payload, DefaultSchema, resolve_relationships=True
                )

        # Patch
        data = {"schema_id": schema.id}
        self._patch(
            path="default_schema", data=data, type="default_schema", obj_id=source.id
        )
        return self._get(
            path="default_schema",
            obj_id=source.id,
            clazz=DefaultSchema,
            resolve_relationships=True,
        )


class Analyze:
    def __init__(self, url: str):
        self._base_url = furl(url) / "api/v1/analyze"
        self._session = requests.Session()

    def analyze(
        self,
        query: str,
        source: Source,
        start_time: datetime.datetime,
        end_time: datetime.datetime,
        name: str = None,
    ) -> JobExecution:
        response = self._session.post(
            self._base_url,
            params={
                "query": query,
                "name": name,
                "source_id": source.id,
                "start_time": start_time.isoformat(),
                "end_time": end_time.isoformat(),
            },
        )
        if response.status_code == 441:
            raise TableNotFound(response.json()["message"])
        elif response.status_code == 442:
            raise ColumnNotFound(response.json()["message"])
        elif response.status_code == 422:
            raise ParseError(response.json()["message"])
        elif response.status_code == 443:
            raise SemanticError(response.json()["message"])

        logging.debug(response.text)
        response.raise_for_status()
        payload = response.json()["data"]
        return JobExecution(
            session=self._session,
            attributes=payload.get("attributes"),
            obj_id=payload.get("id"),
            relationships=None,
        )


class Parse:
    def __init__(self, url: str):
        self._base_url = furl(url) / "api/v1/parse"
        self._session = requests.Session()

    def parse(self, query: str, source: Source):
        response = self._session.post(
            self._base_url, params={"query": query, "source_id": source.id},
        )
        logging.debug(response.text)
        response.raise_for_status()
        return response.json()
